from functools import partial
from typing import Callable, NamedTuple, Optional, Union

import jax
import jax.flatten_util
import jax.numpy as jnp
from jax import grad, jit, lax, random
from jaxopt import OptStep
from jaxopt._src import loop
from jaxopt.prox import prox_none

from ..tree_utils import tree_add_scalar_mul, tree_l2_norm, tree_slice, tree_sub
from ..typing import KeyArrayLike, Pytree


class SVRGState(NamedTuple):
    """
    Optimizer state for (Prox)SVRG.

    Attributes
    ----------
    iter_num :
        Current epoch or iteration number.
    key :
        Random key to use when sampling data points or mini-batches.
    error :
        Scaled difference (~distance) between subsequent parameter values
        used to monitor convergence.
    stepsize :
        Step size of the individual gradient steps.
    reference_point :
        Anchor/reference/snapshot point where the full gradient is calculated in the SVRG algorithm.
        Corresponds to $x_{s}$ in the pseudocode[$^{[1]}$](#references).
    full_grad_at_reference_point :
        Full gradient at the anchor/reference point.

    # References
    ------------
    [1] [Gower, Robert M., Mark Schmidt, Francis Bach, and Peter Richtárik.
        "Variance-Reduced Methods for Machine Learning." arXiv preprint arXiv:2010.00892 (2020).
        ](https://arxiv.org/abs/2010.00892)
    """

    iter_num: int
    key: KeyArrayLike
    error: float
    stepsize: float
    reference_point: Optional[Pytree] = None
    full_grad_at_reference_point: Optional[Pytree] = None


class ProxSVRG:
    """
    Prox-SVRG solver

    Borrowing from jaxopt.ProximalGradient, this solver minimizes:

      objective(params, hyperparams_prox, *args, **kwargs) =
        fun(params, *args, **kwargs) + non_smooth(params, hyperparams_prox)

    Attributes
    ----------
    fun: Callable
        Smooth function of the form ``fun(x, *args, **kwargs)``.
    prox: Callable
        Proximal operator associated with the function ``non_smooth``.
        It should be of the form ``prox(params, hyperparams_prox, scale=1.0)``.
        See ``jaxopt.prox`` for examples.
    maxiter : int
        Maximum number of epochs to run the optimization for.
    key : jax.random.PRNGkey
        jax PRNGKey to start with. Used for sampling random data points.
    stepsize : float
        Constant step size to use.
    tol: float
        Tolerance level for the error when comparing parameters
        at the end of consecutive epochs to check for convergence.
    batch_size: int
        Number of data points to sample per inner loop iteration.

    Examples
    --------
    >>> import numpy as np
    >>> from jaxopt.prox import prox_lasso
    >>> loss_fn = lambda params, X, y: ((X.dot(params) - y)**2).sum()
    >>> svrg = ProxSVRG(loss_fn, prox_lasso)
    >>> hyperparams_prox = 0.1
    >>> params, state = svrg.run(np.zeros(2), hyperparams_prox, np.ones((10, 2)), np.zeros(10))


    References
    ----------
    [1] [Gower, Robert M., Mark Schmidt, Francis Bach, and Peter Richtárik.
    "Variance-Reduced Methods for Machine Learning." arXiv preprint arXiv:2010.00892 (2020).
    ](https://arxiv.org/abs/2010.00892)

    [2] [Xiao, Lin, and Tong Zhang.
    "A proximal stochastic gradient method with progressive variance reduction."
    SIAM Journal on Optimization 24.4 (2014): 2057-2075.](https://arxiv.org/abs/1403.4699v1)

    [3] [Johnson, Rie, and Tong Zhang.
    "Accelerating stochastic gradient descent using predictive variance reduction."
    Advances in neural information processing systems 26 (2013).
    ](https://proceedings.neurips.cc/paper/2013/hash/ac1dd209cbcc5e5d1c6e28598e8cbbe8-Abstract.html)
    """

    def __init__(
        self,
        fun: Callable,
        prox: Callable,
        maxiter: int = 10_000,
        key: Optional[KeyArrayLike] = None,
        stepsize: float = 1e-3,
        tol: float = 1e-3,
        batch_size: int = 1,
    ):
        self.fun = fun
        self.maxiter = maxiter
        self.key = key
        self.stepsize = stepsize
        self.tol = tol
        self.loss_gradient = jit(grad(self.fun))
        self.batch_size = batch_size
        self.proximal_operator = prox

    def init_state(
        self,
        init_params: Pytree,
        *args,
    ) -> SVRGState:
        """
        Initialize the solver state

        Parameters
        ----------
        init_params :
            Pytree containing the initial parameters.
            For GLMs it's a tuple of (W, b)
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        state :
            Initialized optimizer state
        """
        state = SVRGState(
            iter_num=0,
            key=self.key if self.key is not None else random.key(123),
            error=jnp.inf,
            stepsize=self.stepsize,
            reference_point=init_params,
            full_grad_at_reference_point=None,
        )
        return state

    @partial(jit, static_argnums=(0,))
    def _inner_loop_param_update_step(
        self,
        params: Pytree,
        reference_point: Pytree,
        full_grad_at_reference_point: Pytree,
        stepsize: float,
        hyperparams_prox: Union[float, None],
        *args,
    ) -> Pytree:
        """
        Body of the inner loop of Prox-SVRG that takes a step.

        Parameters
        ----------
        params :
            Current parameters.
        reference_point :
            Anchor point.
        full_grad_at_reference_point :
            Full gradient at the anchor point.
        stepsize :
            Step size.
        hyperparams_prox :
            Hyperparameters to `prox`, most commonly regularization strength.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        next_params :
            Parameter values after applying the update.
        """
        # gradient on batch_{i_k} evaluated at the current parameters
        # gradient of f_{i_k} at x_{k} in the pseudocode of Gower et al. 2020
        minibatch_grad_at_current_params = self.loss_gradient(params, *args)
        # gradient on batch_{i_k} evaluated at the anchor point
        # gradient of f_{i_k} at x_{k} in the pseudocode of Gower et al. 2020
        minibatch_grad_at_reference_point = self.loss_gradient(reference_point, *args)

        # SVRG gradient estimate
        gk = jax.tree_util.tree_map(
            lambda a, b, c: a - b + c,
            minibatch_grad_at_current_params,
            minibatch_grad_at_reference_point,
            full_grad_at_reference_point,
        )

        # x_{k+1} = x_{k} - stepsize * g_{k}
        next_params = tree_add_scalar_mul(params, -stepsize, gk)

        # apply the proximal operator
        next_params = self.proximal_operator(
            next_params, hyperparams_prox, scaling=stepsize
        )

        return next_params

    @partial(jit, static_argnums=(0,))
    def update(
        self,
        params: Pytree,
        state: SVRGState,
        hyperparams_prox: Union[float, None],
        *args,
    ) -> OptStep:
        """
        Perform a single parameter update on the passed data (no random sampling or loops)
        and increment `state.iter_num`.

        Please note that this gets called by `BaseRegressor._solver_update` (e.g., as called by `GLM.update`),
        but repeated calls to `(Prox)SVRG.update` (so in turn e.g. to `GLM.update`) on mini-batches passed to it
        will not result in running the full (Prox-)SVRG, and parts of the algorithm will have to be implemented outside.

        Parameters
        ----------
        params :
            Parameters at the end of the previous update, used as the starting point for the current update.
        state :
            Optimizer state at the end of the previous update.
            Needs to have the current anchor point (`reference_point`) and the gradient at the anchor point
            (`full_grad_at_reference_point`) already set.
        hyperparams_prox :
            Hyperparameters to `prox`, most commonly regularization strength.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.


        Returns
        -------
        OptStep
            reference_point :
                Parameters after taking one step defined in the inner loop of Prox-SVRG.
            state :
                Updated state.

        Raises
        ------
        ValueError
            The parameter update needs a value for the full gradient at the anchor point, which needs the full data
            to be calculated and is expected to be stored in `state.full_grad_at_reference_point`. So if
            `state.full_grad_at_reference_point` is None, a ValueError is raised.
        """
        if state.full_grad_at_reference_point is None:
            raise ValueError(
                "Full gradient at the anchor point (state.full_grad_at_reference_point) has to be set."
            )
        return self._update_on_batch(params, state, hyperparams_prox, *args)

    @partial(jit, static_argnums=(0,))
    def _update_on_batch(
        self,
        params: Pytree,
        state: SVRGState,
        hyperparams_prox: Union[float, None],
        *args,
    ) -> OptStep:
        """
        Update parameters given a mini-batch of data and increment iteration/epoch number in state.

        Note that this method doesn't update `state.reference_point`, `state.full_grad_at_reference_point`,
        that has to be done outside.

        Parameters
        ----------
        params :
            Parameters at the end of the previous update, used as the starting point for the current update.
        state :
            Optimizer state at the end of the previous update.
            Needs to have the current anchor point (`reference_point`) and the gradient at the anchor point
            (`full_grad_at_reference_point`) already set.
        hyperparams_prox :
            Hyperparameters to `prox`, most commonly regularization strength.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        OptStep
            reference_point :
                Parameters after taking one step defined in the inner loop of Prox-SVRG.
            state :
                Updated state.
        """
        next_params = self._inner_loop_param_update_step(
            params,
            state.reference_point,
            state.full_grad_at_reference_point,
            state.stepsize,
            hyperparams_prox,
            *args,
        )

        state = state._replace(
            iter_num=state.iter_num + 1,
        )

        return OptStep(params=next_params, state=state)

    @partial(jit, static_argnums=(0,))
    def run(
        self,
        init_params: Pytree,
        hyperparams_prox: Union[float, None],
        *args,
    ) -> OptStep:
        """
        Run a whole optimization until convergence or until `maxiter` epochs are reached.
        Called by `BaseRegressor._solver_run` (e.g. as called by `GLM.fit`) and assumes
        that X and y are the full data set.

        Parameters
        ----------
        init_params :
            Initial parameters to start from.
        hyperparams_prox :
            Hyperparameters to `prox`, most commonly regularization strength.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        OptStep
            final_params :
                Parameters at the end of the last innner loop.
                (... or the average of the parameters over the last inner loop)
            final_state :
                Final optimizer state.
        """
        # initialize the state, including the full gradient at the initial parameters
        init_state = self.init_state(
            init_params,
            *args,
        )

        return self._run(init_params, init_state, hyperparams_prox, *args)

    @partial(jit, static_argnums=(0,))
    def _run(
        self,
        init_params: Pytree,
        init_state: SVRGState,
        hyperparams_prox: Union[float, None],
        *args,
    ) -> OptStep:
        """
        Run a whole optimization until convergence or until `maxiter` epochs are reached.
        Called by `BaseRegressor._solver_run` (e.g. as called by `GLM.fit`) and assumes that
        X and y are the full data set.
        Assumes the state has been initialized, which works a bit differently for SVRG and ProxSVRG.

        Parameters
        ----------
        init_params :
            Initial parameters to start from.
        init_state :
            Initialized optimizer state returned by `ProxSVRG.init_state`
        hyperparams_prox :
            Hyperparameters to `prox`, most commonly regularization strength.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.
        Returns
        -------
        OptStep
            final_params :
                Parameters at the end of the last innner loop.
                (... or the average of the parameters over the last inner loop)
            final_state :
                Final optimizer state.
        """

        # this method assumes that args hold the full data
        def body_fun(step):
            prev_reference_point, state = step

            # evaluate and store the full gradient with the params from the last inner loop
            state = state._replace(
                full_grad_at_reference_point=self.loss_gradient(
                    prev_reference_point, *args
                )
            )

            # run an update over the whole data
            params, state = self._update_per_random_samples(
                prev_reference_point, state, hyperparams_prox, *args
            )

            # update reference point (x_{s}) with the final parameters (x_{m}) or an average over
            # the inner loop's iterations
            # note that the average is currently not implemented
            reference_point = params

            state = state._replace(
                reference_point=reference_point,
                error=self._error(
                    reference_point, prev_reference_point, state.stepsize
                ),
            )

            return OptStep(params=reference_point, state=state)

        # at the end of each epoch, check for convergence or reaching the max number of epochs
        def cond_fun(step):
            _, state = step
            return (state.iter_num <= self.maxiter) & (state.error >= self.tol)

        # initialize the full gradient at the anchor point
        # the anchor point is init_params at first
        init_state = init_state._replace(
            full_grad_at_reference_point=self.loss_gradient(init_params, *args)
        )

        final_params, final_state = loop.while_loop(
            cond_fun=cond_fun,
            body_fun=body_fun,
            init_val=OptStep(params=init_params, state=init_state),
            maxiter=self.maxiter,
            jit=True,
        )
        return OptStep(params=final_params, state=final_state)

    @partial(jit, static_argnums=(0,))
    def _update_per_random_samples(
        self,
        params: Pytree,
        state: SVRGState,
        hyperparams_prox: Union[float, None],
        *args,
    ) -> OptStep:
        """
        Performs the inner loop of Prox-SVRG sweeping through approximately one full epoch,
        updating the parameters after sampling a mini-batch on each iteration.

        Parameters
        ----------
        params :
            Parameters at the end of the previous update, used as the starting point for the current update.
        state :
            Optimizer state at the end of the previous sweep.
            Needs to have the current anchor point (`reference_point`) and the gradient at the anchor point
            (`full_grad_at_reference_point`) already set.
        hyperparams_prox :
            Hyperparameters to `prox`, most commonly regularization strength. Can be None.
        args :
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        OptStep
            next_params :
                Parameters at the end of the last inner loop.
                (... or the average of the parameters over the last inner loop)
            state :
                Updated state.

        Raises
        ------
        ValueError
            If not all arguments in args have the same sized first dimension.
        """
        n_points_per_arg = {leaf.shape[0] for leaf in jax.tree.leaves(args)}
        if not len(n_points_per_arg) == 1:
            raise ValueError("All arguments must have the same sized first dimension.")
        N = n_points_per_arg.pop()

        m = (N + self.batch_size - 1) // self.batch_size  # number of iterations

        def inner_loop_body(_, carry):
            params, key = carry

            # sample mini-batch or data point
            key, subkey = random.split(key)
            ind = random.randint(subkey, (self.batch_size,), 0, N)

            # perform a single update on the mini-batch or data point
            next_params = self._inner_loop_param_update_step(
                params,
                state.reference_point,
                state.full_grad_at_reference_point,
                state.stepsize,
                hyperparams_prox,
                *tree_slice(args, ind),
            )

            return (next_params, key)

        next_params, key = lax.fori_loop(
            0,
            m,
            inner_loop_body,
            (params, state.key),
        )

        # update the state
        # storing the average over the inner loop to potentially use it in the run loop
        state = state._replace(
            iter_num=state.iter_num + 1,
            key=key,
        )

        return OptStep(params=next_params, state=state)

    @staticmethod
    def _error(x, x_prev, stepsize):
        """
        Calculate the magnitude of the update relative to the stepsize.
        Used for terminating the algorithm if a certain tolerance is reached.

        Params
        ------
        x :
            Parameter values after the update.
        x_prev :
            Previous parameter values.

        Returns
        -------
        Scaled update magnitude.
        """
        return tree_l2_norm(tree_sub(x, x_prev)) / stepsize


class SVRG(ProxSVRG):
    """
    SVRG solver.

    This solver implements "Algorithm 3" of [1]. Equivalent to ProxSVRG with prox as the identity
    function and hyperparams_prox=None.

    Attributes
    ----------
    fun: Callable
        smooth function of the form ``fun(x, *args, **kwargs)``.
    maxiter : int
        Maximum number of epochs to run the optimization for.
    key : jax.random.PRNGkey
        jax PRNGKey to start with. Used for sampling random data points.
    stepsize : float
        Constant step size to use.
    tol: float
        Tolerance level for the error when comparing parameters
        at the end of consecutive epochs to check for convergence.
    batch_size: int
        Number of data points to sample per inner loop iteration.

    Examples
    --------
    >>> import numpy as np
    >>> loss_fn = lambda params, X, y: ((X.dot(params) - y)**2).sum()
    >>> svrg = SVRG(loss_fn)
    >>> params, state = svrg.run(np.zeros(2), np.ones((10, 2)), np.zeros(10))

    References
    ----------
    [1] [Gower, Robert M., Mark Schmidt, Francis Bach, and Peter Richtárik.
    "Variance-Reduced Methods for Machine Learning." arXiv preprint arXiv:2010.00892 (2020).
    ](https://arxiv.org/abs/2010.00892)

    [2] [Xiao, Lin, and Tong Zhang. "A proximal stochastic gradient method with progressive variance reduction."
    SIAM Journal on Optimization 24.4 (2014): 2057-2075.](https://arxiv.org/abs/1403.4699v1)

    [3] [Johnson, Rie, and Tong Zhang. "Accelerating stochastic gradient descent using predictive variance reduction."
    Advances in neural information processing systems 26 (2013).
    ](https://proceedings.neurips.cc/paper/2013/hash/ac1dd209cbcc5e5d1c6e28598e8cbbe8-Abstract.html)
    """

    def __init__(
        self,
        fun: Callable,
        maxiter: int = 10_000,
        key: Optional[KeyArrayLike] = None,
        stepsize: float = 1e-3,
        tol: float = 1e-3,
        batch_size: int = 1,
    ):
        super().__init__(
            fun,
            prox_none,
            maxiter,
            key,
            stepsize,
            tol,
            batch_size,
        )

    def init_state(self, init_params: Pytree, *args, **kwargs) -> SVRGState:
        """
        Initialize the solver state

        Parameters
        ----------
        init_params :
            pytree containing the initial parameters.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        state :
            Initialized optimizer state
        """
        return super().init_state(init_params, *args, **kwargs)

    @partial(jit, static_argnums=(0,))
    def update(self, params: Pytree, state: SVRGState, *args, **kwargs) -> OptStep:
        """
        Perform a single parameter update on the passed data (no random sampling or loops)
        and increment `state.iter_num`.

        Please note that this gets called by `BaseRegressor._solver_update` (e.g., as called by `GLM.update`),
        but repeated calls to `(Prox)SVRG.update` (so in turn e.g. to `GLM.update`) on mini-batches passed to it
        will not result in running the full (Prox-)SVRG, and parts of the algorithm will have to be implemented outside.

        Parameters
        ----------
        params :
            Parameters at the end of the previous update, used as the starting point for the current update.
        state :
            Optimizer state at the end of the previous update.
            Needs to have the current anchor point (`reference_point`) and the gradient at the anchor point
            (`full_grad_at_reference_point`) already set.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        OptStep
            reference_point :
                Parameters after taking one step defined in the inner loop of Prox-SVRG.
            state :
                Updated state.

        Raises
        ------
        ValueError
            The parameter update needs a value for the full gradient at the anchor point, which needs the full data
            to be calculated and is expected to be stored in `state.full_grad_at_reference_point`.
            So if `state.full_grad_at_reference_point` is None, a ValueError is raised.
        """
        # substitute None for hyperparams_prox
        return super().update(params, state, None, *args, **kwargs)

    @partial(jit, static_argnums=(0,))
    def run(
        self,
        init_params: Pytree,
        *args,
    ) -> OptStep:
        """
        Run a whole optimization until convergence or until `maxiter` epochs are reached.
        Called by `BaseRegressor._solver_run` (e.g. as called by `GLM.fit`) and assumes that
        X and y are the full data set.

        Parameters
        ----------
        init_params :
            Initial parameters to start from.
        args:
            Positional arguments passed to loss function `fun` and its gradient (e.g. `fun(params, *args)`),
            most likely input and output data.
            They are expected to be Pytrees with arrays or FeaturePytree as their leaves, with all of their
            leaves having the same sized first dimension (corresponding to the number of data points).
            For GLMs these are:
                X : DESIGN_INPUT_TYPE
                    Input data.
                y : jnp.ndarray
                    Output data.

        Returns
        -------
        OptStep
            final_params :
                Parameters at the end of the last innner loop.
                (... or the average of the parameters over the last inner loop)
            final_state :
                Final optimizer state.
        """
        # initialize the state, including the full gradient at the initial parameters
        # don't have to pass hyperparams_prox here
        init_state = self.init_state(init_params, *args)

        # substitute None for hyperparams_prox
        return self._run(init_params, init_state, None, *args)



from contextlib import nullcontext as does_not_raise

import jax.numpy as jnp
import numpy as np
import pynapple as nap
import pytest

import nemos as nmo


@pytest.mark.parametrize(
    "inp, jax_func, expected_type",
    [
        ([jnp.arange(3)], lambda x: jnp.power(x, 2), jnp.ndarray),
        ([np.arange(3)], lambda x: jnp.power(x, 2), jnp.ndarray),
        ([nap.Tsd(t=np.arange(3), d=np.arange(3))], lambda x: jnp.power(x, 2), nap.Tsd),
        (
            [nap.TsdFrame(t=np.arange(3), d=np.expand_dims(np.arange(3), 1))],
            lambda x: jnp.power(x, 2),
            nap.TsdFrame,
        ),
        (
            [nap.TsdTensor(t=np.arange(3), d=np.expand_dims(np.arange(3), (1, 2, 3)))],
            lambda x: jnp.power(x, 2),
            nap.TsdTensor,
        ),
        ([nmo.glm.GLM()], lambda x: x, nmo.glm.GLM),
        (
            [nap.Tsd(t=np.arange(3), d=np.arange(3)), np.arange(3)],
            lambda x, y: (x + y, y),
            (nap.Tsd, nap.Tsd),
        ),
        (
            [nap.Tsd(t=np.arange(3), d=np.arange(3)), np.expand_dims(np.arange(3), 1)],
            lambda x, y: (np.power(x, 2), x + y),
            (nap.Tsd, nap.TsdFrame),
        ),
        (
            [
                nap.Tsd(t=np.arange(3), d=np.arange(3)),
                np.expand_dims(np.arange(3), (1, 2)),
            ],
            lambda x, y: (np.power(x, 2), x + y),
            (nap.Tsd, nap.TsdTensor),
        ),
        # Test adding a scalar to different types of inputs
        ([np.arange(3)], lambda x: x + 5, np.ndarray),
        ([jnp.arange(3)], lambda x: x + 5, jnp.ndarray),
        ([nap.Tsd(t=np.arange(3), d=np.arange(3))], lambda x: x + 5, nap.Tsd),
        # Test element-wise multiplication between array and Tsd
        (
            [np.arange(3), nap.Tsd(t=np.arange(3), d=np.arange(3))],
            lambda x, y: x * y,
            nap.Tsd,
        ),
        # Test concatenation of TsdFrames along a new dimension
        (
            [
                nap.TsdFrame(t=np.arange(3), d=np.arange(3).reshape(-1, 1)),
                nap.TsdFrame(t=np.arange(3), d=np.arange(3, 6).reshape(-1, 1)),
            ],
            lambda x, y: np.concatenate([x, y], axis=1),
            nap.TsdFrame,
        ),
        # Test operation that reduces dimensionality, from TsdTensor to TsdFrame
        (
            [nap.TsdTensor(t=np.arange(3), d=np.random.rand(3, 2, 2))],
            lambda x: x.mean(axis=2),  # Reduce last dimension
            nap.TsdFrame,
        ),
        # Test mixing JAX array and NumPy array inputs resulting in a JAX array
        (
            [jnp.arange(3), np.arange(3)],
            lambda x, y: x + y,
            jnp.ndarray,
        ),
        # Test operation involving three inputs of mixed types
        (
            [
                np.arange(3),
                jnp.arange(3, 6),
                nap.Tsd(t=np.arange(6, 9), d=np.arange(6, 9)),
            ],
            lambda x, y, z: (x + y, y + z),
            (nap.Tsd, nap.Tsd),
        ),
    ],
)
def test_decorator_output_type(inp, jax_func, expected_type):
    """Validate that the `cast_jax` decorator correctly casts output types based on input types."""

    @nmo.type_casting.support_pynapple(conv_type="jax")
    def func(*x):
        return jax_func(*x)

    out = func(*inp)
    assert nmo.tree_utils.pytree_map_and_reduce(
        lambda x, y: isinstance(x, y), all, out, expected_type
    )


@pytest.mark.parametrize(
    "inp, jax_func, expected_type",
    [
        ([jnp.arange(3)], lambda x: jnp.power(x, 2), jnp.ndarray),
        ([np.arange(3)], lambda x: np.power(x, 2), np.ndarray),
        ([nap.Tsd(t=np.arange(3), d=np.arange(3))], lambda x: jnp.power(x, 2), nap.Tsd),
        (
            [nap.TsdFrame(t=np.arange(3), d=np.expand_dims(np.arange(3), 1))],
            lambda x: jnp.power(x, 2),
            nap.TsdFrame,
        ),
        (
            [nap.TsdTensor(t=np.arange(3), d=np.expand_dims(np.arange(3), (1, 2, 3)))],
            lambda x: jnp.power(x, 2),
            nap.TsdTensor,
        ),
        ([nmo.glm.GLM()], lambda x: x, nmo.glm.GLM),
        (
            [nap.Tsd(t=np.arange(3), d=np.arange(3)), np.arange(3)],
            lambda x, y: (x + y, y),
            (nap.Tsd, nap.Tsd),
        ),
        (
            [nap.Tsd(t=np.arange(3), d=np.arange(3)), np.expand_dims(np.arange(3), 1)],
            lambda x, y: (np.power(x, 2), x + y),
            (nap.Tsd, nap.TsdFrame),
        ),
        (
            [
                nap.Tsd(t=np.arange(3), d=np.arange(3)),
                np.expand_dims(np.arange(3), (1, 2)),
            ],
            lambda x, y: (np.power(x, 2), x + y),
            (nap.Tsd, nap.TsdTensor),
        ),
        # Test adding a scalar to different types of inputs
        ([np.arange(3)], lambda x: x + 5, np.ndarray),
        ([jnp.arange(3)], lambda x: x + 5, jnp.ndarray),
        ([nap.Tsd(t=np.arange(3), d=np.arange(3))], lambda x: x + 5, nap.Tsd),
        # Test element-wise multiplication between array and Tsd
        (
            [np.arange(3), nap.Tsd(t=np.arange(3), d=np.arange(3))],
            lambda x, y: x * y,
            nap.Tsd,
        ),
        # Test concatenation of TsdFrames along a new dimension
        (
            [
                nap.TsdFrame(t=np.arange(3), d=np.arange(3).reshape(-1, 1)),
                nap.TsdFrame(t=np.arange(3), d=np.arange(3, 6).reshape(-1, 1)),
            ],
            lambda x, y: np.concatenate([x, y], axis=1),
            nap.TsdFrame,
        ),
        # Test operation that reduces dimensionality, from TsdTensor to TsdFrame
        (
            [nap.TsdTensor(t=np.arange(3), d=np.random.rand(3, 2, 2))],
            lambda x: x.mean(axis=2),  # Reduce last dimension
            nap.TsdFrame,
        ),
        # Test mixing JAX array and NumPy array inputs resulting in a JAX array
        (
            [jnp.arange(3), np.arange(3)],
            lambda x, y: x + y,
            jnp.ndarray,
        ),
        # Test operation involving three inputs of mixed types
        (
            [
                np.arange(3),
                jnp.arange(3, 6),
                nap.Tsd(t=np.arange(6, 9), d=np.arange(6, 9)),
            ],
            lambda x, y, z: (x + y, y + z),
            (nap.Tsd, nap.Tsd),
        ),
    ],
)
def test_decorator_output_type_numpy(inp, jax_func, expected_type):
    """Validate that the `cast_jax` decorator correctly casts output types based on input types."""

    @nmo.type_casting.support_pynapple(conv_type="numpy")
    def func(*x):
        return jax_func(*x)

    out = func(*inp)
    assert nmo.tree_utils.pytree_map_and_reduce(
        lambda x, y: isinstance(x, y), all, out, expected_type
    )


@pytest.mark.parametrize(
    "iset",
    [
        nap.IntervalSet(start=[0], end=[1]),
        nap.IntervalSet(start=[0, 0.5], end=[0.49, 1]),
    ],
)
def test_predict_nap_output(iset, poissonGLM_model_instantiation):
    """Ensure predictions using pynapple objects correctly respect interval sets."""
    X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
    n_samp = X.shape[0]
    time = jnp.linspace(0, 1, n_samp)
    tsd_X = nap.TsdFrame(t=time, d=X).restrict(iset)
    model.fit(X, y)
    # run predict on tsd
    pred_rate = model.predict(tsd_X)
    assert isinstance(pred_rate, nap.Tsd)
    assert np.all(iset.values == pred_rate.time_support.values)


@pytest.mark.parametrize(
    "inp, expected",
    [
        (np.array([1, 2, 3]), True),
        (jnp.array([1, 2, 3]), True),
        ([1, 2, 3], False),
        (3, False),
        ("not an array", False),
        (
            nap.Tsd(t=np.arange(3), d=np.arange(3)),
            True,
        ),  # Assuming Tsd objects have array-like properties
    ],
)
def test_is_numpy_array_like(inp, expected):
    """Check if various objects are correctly identified as numpy array-like or not."""
    assert nmo.type_casting.is_numpy_array_like(inp) == expected


@pytest.mark.parametrize(
    "inp, expected",
    [
        (nap.Tsd(t=np.arange(3), d=np.arange(3)), True),
        (nap.TsdFrame(t=np.arange(3), d=np.arange(3).reshape(-1, 1)), True),
        (nap.TsdTensor(t=np.arange(3), d=np.arange(3).reshape(-1, 1, 1)), True),
        (np.array([1, 2, 3]), False),
        ([1, 2, 3], False),
    ],
)
def test_is_pynapple_tsd(inp, expected):
    """Verify identification of pynapple Tsd/TsdFrame/TsdTensor objects amongst inputs."""
    assert nmo.type_casting.is_pynapple_tsd(inp) == expected


@pytest.mark.parametrize(
    "inputs, expected",
    [
        (
            [
                nap.Tsd(t=np.arange(3), d=np.arange(3)),
                nap.Tsd(t=np.arange(3), d=np.arange(3)),
            ],
            True,
        ),
        (
            [
                nap.Tsd(t=np.arange(3), d=np.arange(3)),
                nap.Tsd(t=np.arange(3, 6), d=np.arange(3)),
            ],
            False,
        ),
    ],
)
def test_all_same_time_info(inputs, expected):
    """Assess if provided pynapple objects share consistent time axis and support information."""
    assert nmo.type_casting.all_same_time_info(*inputs) == expected


@pytest.mark.parametrize(
    "array, time, time_support, expected_type",
    [
        (np.random.rand(3), np.arange(3), nap.IntervalSet(start=[0], end=[3]), nap.Tsd),
        (
            np.random.rand(3, 1),
            np.arange(3),
            nap.IntervalSet(start=[0], end=[3]),
            nap.TsdFrame,
        ),
        (
            np.random.rand(3, 2, 2),
            np.arange(3),
            nap.IntervalSet(start=[0], end=[3]),
            nap.TsdTensor,
        ),
    ],
)
def test_cast_to_pynapple(array, time, time_support, expected_type):
    """Check the conversion from array to the correct pynapple object type based on dimensions."""
    result = nmo.type_casting.cast_to_pynapple(array, time, time_support)
    assert isinstance(result, expected_type)


@pytest.mark.parametrize(
    "inp, expected",
    [
        (np.array([1, 2, 3]), jnp.ndarray),
        ([1, 2, 3], list),  # Assuming list should be converted
        ("not an array", str),  # Assuming strings are not converted
    ],
)
def test_jnp_asarray_if(inp, expected):
    """Evaluate conditional conversion to JAX array based on input characteristics."""
    result = nmo.type_casting.jnp_asarray_if(inp)
    assert isinstance(result, expected)


@pytest.mark.parametrize(
    "inp, expected",
    [
        ([], True),
        ([jnp.array([0, 1]), jnp.array([0, 1])], True),
        ([jnp.array([0, 1]), jnp.array([0, 1]), jnp.array([0, 1])], True),
        ([jnp.array([0, 1]), jnp.array([0, 2]), jnp.array([0, 1])], False),
        ([np.array([0, 1]), jnp.array([0, 1]), jnp.array([0, 1])], True),
        ([jnp.array([0, 1]), jnp.array([0, 1, 2])], False),
        ([jnp.array([[0, 1]]), jnp.array([0, 1])], False),
    ],
)
def test_check_all_close(inp, expected):
    """Evaluate conditional conversion to JAX array based on input characteristics."""
    assert nmo.type_casting._check_all_close(inp) == expected


@pytest.mark.parametrize(
    "data, cls",
    [
        (np.zeros((10,)), nap.Tsd),
        (np.zeros((10, 1)), nap.TsdFrame),
        (np.zeros((10, 1, 1)), nap.TsdTensor),
    ],
)
@pytest.mark.parametrize(
    "t1, t2, expectation",
    [
        (np.arange(10), np.arange(10), does_not_raise()),
        (
            np.arange(10),
            np.arange(10) + 1,
            pytest.raises(
                ValueError,
                match="Time axis mismatch. pynapple objects have mismatching",
            ),
        ),
    ],
)
def test_equal_time_axis_nap_types(t1, t2, data, cls, expectation):
    @nmo.type_casting.support_pynapple(conv_type="jax")
    def func(*x):
        return x

    with expectation:
        func(cls(t=t1, d=data), cls(t=t2, d=data))


@pytest.mark.parametrize(
    "tsds, expectation",
    [
        (
            [
                nap.Tsd(t=np.arange(10), d=np.arange(10)),
                nap.Tsd(t=np.arange(11), d=np.arange(11)),
            ],
            pytest.raises(
                ValueError,
                match="Time axis mismatch. pynapple objects have mismatching",
            ),
        ),
        (
            [
                nap.Tsd(t=np.arange(10), d=np.arange(10)),
                nap.Tsd(
                    t=np.arange(1), d=np.arange(1), time_support=nap.IntervalSet(0, 10)
                ),
                nap.Tsd(t=np.arange(10), d=np.arange(10)),
            ],
            pytest.raises(
                ValueError,
                match="Time axis mismatch. pynapple objects have mismatching",
            ),
        ),
    ],
)
@pytest.mark.parametrize("conv_type", ["numpy", "jax"])
def test_equal_time_axis_different_len(tsds, conv_type, expectation):
    @nmo.type_casting.support_pynapple(conv_type=conv_type)
    def func(*x):
        return x

    with expectation:
        func(*tsds)


@pytest.mark.parametrize(
    "conv_type, expectation",
    [
        ("numpy", does_not_raise()),
        ("jax", does_not_raise()),
        (
            "not_implemented",
            pytest.raises(
                NotImplementedError, match="Conversion of type 'not_implemented'"
            ),
        ),
    ],
)
def test_conv_type(conv_type, expectation):
    @nmo.type_casting.support_pynapple(conv_type=conv_type)
    def func(*x):
        return x

    with expectation:
        func(nap.Tsd(t=np.arange(10), d=np.arange(10)))


@pytest.mark.parametrize("conv_type", ["jax", "numpy"])
def test_time_axis_change_len(conv_type):
    @nmo.type_casting.support_pynapple(conv_type=conv_type)
    def change_shape(x):
        return x[:-1]

    out = change_shape(nap.Tsd(t=np.arange(10), d=np.arange(10)))
    assert isinstance(out, (np.ndarray, jnp.ndarray))

    @nmo.type_casting.support_pynapple(conv_type=conv_type)
    def same_shape(x):
        return x

    out = same_shape(nap.Tsd(t=np.arange(10), d=np.arange(10)))
    assert isinstance(out, nap.Tsd)



"""Abstract class for estimators."""

# required to get ArrayLike to render correctly
from __future__ import annotations

import inspect
import warnings
from collections import defaultdict
from typing import Any


class Base:
    """Base class for NeMoS estimators.

    A base class for estimators with utilities for getting and setting parameters,
    and for interacting with specific devices like CPU, GPU, and TPU.

    This class provides utilities for:
    - Getting and setting parameters using introspection.
    - Sending arrays to target devices (CPU, GPU, TPU).

    Notes
    -----
    The class provides helper methods mimicking scikit-learn's get_params and set_params.
    Additionally, it has methods for selecting target devices and sending arrays to them.
    """

    def get_params(self, deep=True) -> dict:
        """
        From scikit-learn, get parameters by inspecting init.

        Parameters
        ----------
        deep

        Returns
        -------
            out:
                A dictionary containing the parameters. Key is the parameter
                name, value is the parameter value.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            if deep and hasattr(value, "get_params") and not isinstance(value, type):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params: Any):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as :class:`~sklearn.pipeline.Pipeline`). The latter have
        parameters of the form ``<component>__<parameter>`` so that it's
        possible to update each component of a nested object.

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self : estimator instance
            Estimator instance.
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)
        nested_params: defaultdict = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                local_valid_params = self._get_param_names()
                raise ValueError(
                    f"Invalid parameter {key!r} for estimator {self}. "
                    f"Valid parameters are: {local_valid_params!r}."
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            # TODO(1.4): remove specific handling of "base_estimator".
            # The "base_estimator" key is special. It was deprecated and
            # renamed to "estimator" for several estimators. This means we
            # need to translate it here and set sub-parameters on "estimator",
            # but only if the user did not explicitly set a value for
            # "base_estimator".
            if (
                key == "base_estimator"
                and valid_params[key] == "deprecated"
                and self.__module__.startswith("sklearn.")
            ):
                warnings.warn(
                    (
                        f"Parameter 'base_estimator' of {self.__class__.__name__} is"
                        " deprecated in favor of 'estimator'. See"
                        f" {self.__class__.__name__}'s docstring for more details."
                    ),
                    FutureWarning,
                    stacklevel=2,
                )
                key = "estimator"
            valid_params[key].set_params(**sub_params)

        return self

    @classmethod
    def _get_param_names(cls):
        """Get parameter names for the estimator."""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [
            p
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "GLM estimators should always "
                    "specify their parameters in the signature"
                    " of their __init__ (no varargs)."
                    " %s with constructor %s doesn't "
                    " follow this convention." % (cls, init_signature)
                )

        # Consider the constructor parameters excluding 'self'
        parameters = [
            p.name for p in init_signature.parameters.values() if p.name != "self"
        ]

        # remove kwargs
        if "kwargs" in parameters:
            parameters.remove("kwargs")
        # Extract and sort argument names excluding 'self'
        return sorted(parameters)



import jax
import jax.numpy as jnp
import numpy as np
import pytest

import nemos as nmo
from nemos.pytrees import FeaturePytree


class TestFeaturePytree:

    def test_key_error_init(self):
        """Ensure TypeError is raised if non-string keys are used during initialization."""
        with pytest.raises(TypeError, match="keywords must be strings"):
            FeaturePytree(**{1: np.random.rand(100)})

    def test_key_error_set(self):
        """Validate that setting a non-string key results in a ValueError."""
        tree = FeaturePytree(**{"test": np.random.rand(100)})
        with pytest.raises(ValueError, match="Keys must be strings"):
            tree[1] = np.random.rand(100)

    def test_num_timepoints_error_init(self):
        """Check for ValueError if arrays with different time points are initialized."""
        with pytest.raises(
            ValueError, match="All arrays must have same number of time points"
        ):
            FeaturePytree(test1=np.random.rand(100, 1), test2=np.random.rand(50, 1))

    def test_num_timepoints_error_set(self):
        """Ensure setting items with different number of time points raises ValueError."""
        tree = FeaturePytree(test1=np.random.rand(100, 1))
        with pytest.raises(
            ValueError, match="All arrays must have same number of time points"
        ):
            tree["test2"] = np.random.rand(50, 1)

    def test_array_error_init(self):
        """Validate that initializing with non-array values raises a ValueError."""
        with pytest.raises(ValueError, match="All values must be arrays"):
            FeaturePytree(test1="hi")

    def test_array_error_set(self):
        """Ensure setting non-array values results in a ValueError."""
        tree = FeaturePytree(test1=np.random.rand(100, 1))
        with pytest.raises(ValueError, match="All values must be arrays"):
            tree["test2"] = "hi"

    def test_diff_shapes(self):
        """Test handling of arrays with different shapes but the same length."""
        tree = FeaturePytree(test=np.random.rand(100))
        for dim in [1, 2, 3, 4]:
            tree[f"test{dim}"] = np.random.rand(100, dim)
        assert len(tree) == 100

    def test_diff_dims(self):
        """Check for correct handling of arrays with different dimensions but the same length."""
        tree = FeaturePytree(test=np.random.rand(100))
        for ndim in [1, 2, 3, 4]:
            tree[f"test{ndim}"] = np.random.rand(100, *[1] * ndim)
        assert len(tree) == 100

    def test_treemap(self):
        """Test the application of jax.tree_util.tree_map function on FeaturePytree."""
        tree = FeaturePytree(test=np.random.rand(100, 1), test2=np.random.rand(100, 2))
        mapped = jax.tree_util.tree_map(lambda x: jnp.mean(x, axis=-1), tree)
        assert len(tree) == len(mapped)
        assert list(tree.keys()) == list(mapped.keys())
        assert isinstance(mapped, FeaturePytree)

    def test_treemap_npts(self):
        """Check if jax.tree_util.tree_map correctly modifies the number of points in FeaturePytree."""
        tree = FeaturePytree(test=np.random.rand(100, 1), test2=np.random.rand(100, 2))
        mapped = jax.tree_util.tree_map(lambda x: x[::10], tree)
        assert len(mapped) == 10
        assert list(tree.keys()) == list(mapped.keys())

    def test_treemap_to_dict(self):
        """Ensure jax.tree_util.tree_map transforms FeaturePytree to a dictionary with mean values."""
        tree = FeaturePytree(
            test=np.random.rand(
                100,
            ),
            test2=np.random.rand(100, 2),
        )

        mapped = jax.tree_util.tree_map(jnp.mean, tree)
        assert isinstance(mapped, dict)
        assert list(tree.keys()) == list(mapped.keys())

    def test_get_key(self):
        """Test retrieving an item with its key and handling of non-existent keys."""
        test = np.random.rand(100, 1)
        tree = FeaturePytree(test=test)
        np.testing.assert_equal(tree["test"], test)
        with pytest.raises(KeyError):
            tree["hi"]

    def test_get_slice(self):
        """Check slicing functionality of FeaturePytree."""
        tree = FeaturePytree(test=np.random.rand(100, 1), test2=np.random.rand(100, 2))
        assert len(tree[:10]) == 10
        assert list(tree.keys()) == list(tree[:10].keys())
        assert all(tree[:10]["test"] == tree["test"][:10])

    def test_glm(self):
        """Validate the Generalized Linear Model (GLM) implementation with FeaturePytree."""
        w_true = dict(test=np.random.rand(3), test2=np.random.rand(2))
        X = FeaturePytree(test=np.random.rand(100, 3), test2=np.random.rand(100, 2))
        rate = nmo.tree_utils.pytree_map_and_reduce(
            lambda w, x: jnp.einsum("k,tk->t", w, x), sum, w_true, X.data
        )
        spikes = np.random.poisson(rate)
        model = nmo.glm.GLM()
        model.fit(X, spikes)
        assert list(model.coef_.keys()) == list(X.keys())
        for k in model.coef_.keys():
            assert model.coef_[k].shape == X[k].shape[1:]

    def test_eq(self):
        """Check that we can test equality."""
        tree1 = FeaturePytree(test=np.arange(100))
        tree2 = FeaturePytree(test=np.arange(100))
        assert tree1 == tree2

    def test_eq_fail_keys(self):
        """Check that equality fails when keys differ."""
        tree1 = FeaturePytree(test=np.arange(100))
        tree2 = FeaturePytree(test2=np.arange(100))
        assert tree1 != tree2

    def test_eq_fail_shape(self):
        """Check that equality fails when shapes differ."""
        tree1 = FeaturePytree(test=np.arange(100))
        tree2 = FeaturePytree(test=np.arange(50))
        assert tree1 != tree2

    def test_eq_fail_values(self):
        """Check that equality fails when values differ."""
        tree1 = FeaturePytree(test=np.random.rand(100))
        tree2 = FeaturePytree(test=np.random.rand(100))
        assert tree1 != tree2



"""Convolution utilities."""

# required to get ArrayLike to render correctly
from __future__ import annotations

import re
import warnings
from functools import partial
from typing import Any, Literal, Optional

import jax
import jax.numpy as jnp
from numpy.typing import ArrayLike, NDArray

from . import type_casting, utils

_CORR_VEC_BASIS = jax.vmap(partial(jnp.convolve, mode="valid"), (None, 1), 1)

_CORR_VEC = jax.vmap(partial(jnp.convolve, mode="valid"), (1, None), 1)
_CORR_VEC = jax.vmap(_CORR_VEC, (None, 1), 2)


@jax.jit
def tensor_convolve(array: NDArray, eval_basis: NDArray):
    """
    Apply a convolution on the given array with the evaluation basis and reshapes the result.

    This function first flattens the input array across dimensions other than the first one, then
    performs a vectorized convolution with the evaluation basis. The result is reshaped back to the
    original array's shape (except for the first dimension, which is adjusted based on the convolution).

    Parameters
    ----------
    array :
        The input array to convolve. It is expected to be at least 1D. The first axis is expeted to be
        the sample axis, i.e. the shape of array is ``(num_samples, ...)``.
    eval_basis :
        The evaluation basis array for convolution. It should be 2D, where the first dimension
        represents the window size for convolution. Shape ``(window_size, n_basis_funcs)``.

    Returns
    -------
    :
        The convolved array, reshaped to maintain the original dimensions except for the first one,
        which is adjusted based on the window size of ``eval_basis``.

    Notes
    -----
    The convolution implemented here is in mode ``"valid"``. This implies that the time axis shrinks
    ``num_samples - window_size + 1``, where num_samples is the first size of the first axis of ``array``
    and ``window_size`` is the size of the first axis in ``eval_basis``.
    """
    # flatten over other dims & apply vectorized conv
    conv = _CORR_VEC(array.reshape(array.shape[0], -1), eval_basis)

    # unravel the dimensions
    window_size = eval_basis.shape[0]
    num_samples = array.shape[0]
    conv = conv.reshape(
        num_samples - window_size + 1, *array.shape[1:], eval_basis.shape[1]
    )
    return conv


def _shift_time_axis_and_convolve(array: NDArray, eval_basis: NDArray, axis: int):
    """
    Shifts the specified axis to the first position, applies convolution, and then reverses the shift.

    This applies a convolution along a specific axis of a multi-dimensional array. The process involves
    three steps: shifting the axis, convolving, and then reversing the shift to maintain the original
    axis order.

    Parameters
    ----------
    array : NDArray
        The input array for convolution.
    eval_basis : NDArray
        The evaluation basis array for convolution, should be 2D.
    axis : int
        The axis along which the convolution is applied. This axis is temporarily shifted
        to the first position for the convolution operation.

    Returns
    -------
    NDArray
        The array after applying the convolution along the specified axis, with the original
        axis order restored.

    Notes
    -----
    This function supports arrays of any dimensionality greater or equal than 1.
    """
    # move time axis to first
    new_axis = (jnp.arange(array.ndim) + axis) % array.ndim
    array = jnp.transpose(array, new_axis)

    # convolve
    if array.ndim > 1:
        conv = tensor_convolve(array, eval_basis)
    else:
        conv = _CORR_VEC_BASIS(array, eval_basis)

    # reverse transposition
    new_axis = (*((jnp.arange(array.ndim) - axis) % array.ndim), array.ndim)
    conv = jnp.transpose(conv, new_axis)
    return conv


def _list_epochs(tsd: Any):
    """List epochs from a time series with data object, supporting 'pynapple' Tsd formats.

    If the input is recognized as a 'pynapple' TSD, it extracts epochs based on the TSD's
    time support. Otherwise, it returns the input as is, assuming it's a single epoch.

    Parameters
    ----------
    tsd :
        The time series data, potentially in Pynapple TSD format or a generic data structure.

    Returns
    -------
    :
        A list of epochs extracted from the TSD. If the input is not a Pynapple TSD, returns
        a list containing the input itself.
    """
    if type_casting.is_pynapple_tsd(tsd):
        return [tsd.get(s, e) for s, e in tsd.time_support.values]
    return [tsd]


def _convolve_pad_and_shift(
    basis_matrix: ArrayLike,
    time_series: Any,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    axis: int = 0,
    shift: Optional[bool] = None,
):
    """Create predictor by convolving basis_matrix with time_series.

    To create the convolutional predictor, three steps are taken, each of
    calls a single function. See their docstrings for more details about
    each step. This function **preserves** the number of samples by
    NaN-padding appropriately.

    - Convolve `basis_matrix` with `time_series` (function: `convolve_1d_trials`)
    - Pad output with `basis_matrix.shape[0]-1` NaNs, with location based on
      causality of intended predictor (function: `nan_pad`).
    - (Optional) Shift predictor based on causality (function: `shift_time_series`)

    Parameters
    ----------
    basis_matrix :
        The basis matrix with which to convolve the trials. Shape
        `(window_size, n_basis_funcs)`.
    time_series :
        The time series to convolve with the basis matrix. This variable should
        be a pytree with arrays of at least one-dimension as leaves.
    predictor_causality:
        Causality of this predictor, which determines where padded values are
        added and how the predictor is shifted.
    axis:
        Axis containing samples.
    shift :
        Whether to shift predictor based on causality (only valid if
        `predictor_causality != 'acausal'`). Default is True for `causal` and
        `anti-causal`, False for `acausal`.

    Returns
    -------
    predictor :
        Predictor of with same shape and structure as `time_series`
    """

    # apply convolution
    def conv(x):
        return _shift_time_axis_and_convolve(x, basis_matrix, axis=axis)

    predictor = jax.tree_util.tree_map(conv, time_series)

    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")
        predictor = utils.nan_pad(
            predictor, basis_matrix.shape[0] - 1, predictor_causality, axis=axis
        )

    for w in warns:
        if re.match("^With acausal filter", str(w.message)):
            warnings.warn(
                message="With `acausal` filter, `basis_matrix.shape[0] "
                "should probably be odd, so that we can place an equal number of NaNs on "
                "either side of input.",
                category=UserWarning,
            )
        else:
            warnings.warn(w.message, w.category)

    if shift:
        predictor = utils.shift_time_series(predictor, predictor_causality, axis=axis)
    return predictor


def create_convolutional_predictor(
    basis_matrix: ArrayLike,
    time_series: Any,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    shift: Optional[bool] = None,
    axis: int = 0,
):
    """Create a convolutional predictor by convolving a basis matrix with a time series.

    To create the convolutional predictor, three steps are taken, convolve, pad and shift.

    - Convolve `basis_matrix` with `time_series` (function: `_convolve_1d_trials`)
    - Pad output with `basis_matrix.shape[0]-1` NaNs, with location based on
      causality of intended predictor (function: `nemos.utils.nan_pad`).
    - (Optional) Shift predictor based on causality (function: `nemos.utils.shift_time_series`)

    The function is designed to handle both single arrays and pynapple time series with data
     (across multiple epochs), as well as their combinations. For pynapple time series, it
    treats each epoch separately, applies the convolution process, and then reassembles the time series.

    Parameters
    ----------
    basis_matrix :
        A 2D array representing the basis matrix to convolve with the time series.
        The first dimension should represent the window size for the convolution.
    time_series :
        The time series data to convolve with the basis matrix. Can be single arrays, pytree of arrays,
        pynapple time series, pytree of pynapple time series or a mix.
        In case of Pynapple time series data, each epoch will be convolved separately.
    predictor_causality :
        The causality of the predictor, determining how the padding and shifting
        should be applied to the convolution result.
        - 'causal': Pads and/or shifts the result to be causal with respect to the input.
        - 'acausal': Applies padding equally on both sides without shifting.
        - 'anti-causal': Pads and/or shifts the result to be anti-causal with respect to the input.
    shift :
        Determines whether to shift the convolution result based on the causality.
        If None, it defaults to True for 'causal' and 'anti-causal' and to False for 'acausal'.
    axis :
        The axis along which the convolution is applied.

    Returns
    -------
    Any
        The convolutional predictor, structured similarly to the input `time_series`, appropriately
        padded and shifted according to the specified causality.

    Raises
    ------
    ValueError:
        If `basis_matrix` is not a 2-dimensional array or has a singleton first dimension.
    ValueError:
        If `time_series` does not contain arrays of at least one dimension or contains
        arrays with a dimensionality less than `axis`.
    ValueError:
        If any array within `time_series` or `basis_matrix` is empty.
    ValueError:
        If the number of elements along the convolution axis in any array within `time_series`
        is less than the window size of the `basis_matrix`.
    ValueError:
        If shifting is attempted with 'acausal' causality.
    """
    # convert to jnp.ndarray
    basis_matrix = jnp.asarray(basis_matrix)
    if not utils.check_dimensionality(basis_matrix, 2):
        raise ValueError(
            "basis_matrix must be a 2 dimensional array! "
            f"{basis_matrix.ndim} dimensions provided instead."
        )

    if basis_matrix.shape[0] == 1:
        raise ValueError("`basis_matrix.shape[0]` should be at least 2!")

    # check for empty inputs
    utils.check_non_empty(basis_matrix, "basis_matrix")
    utils.check_non_empty(time_series, "time_series")

    # check sample_axis exists
    if not utils.pytree_map_and_reduce(lambda x: x.ndim > axis, all, time_series):
        raise ValueError(
            "`time_series` should contain arrays of at least one-dimension. "
            "At list one 0-dimensional array provided."
        )

    # assign defaults
    if shift is None:
        if predictor_causality == "acausal":
            shift = False
        else:
            shift = True

    if shift and predictor_causality == "acausal":
        raise ValueError(
            "Cannot shift `predictor` when `predictor_causality` is `acausal`!"
        )

    # flatten and grab tree struct
    flat_tree, struct = jax.tree_util.tree_flatten(time_series)

    # find pynapple
    is_nap = list(type_casting.is_pynapple_tsd(x) for x in flat_tree)

    # retrieve time info
    time_info = [
        type_casting.get_time_info(ts) if is_nap[i] else None
        for i, ts in enumerate(flat_tree)
    ]

    # split epochs (adds one layer to pytree)
    two_layer = jax.tree_util.tree_map(_list_epochs, flat_tree)

    # check trial size (after splitting)
    utils.check_trials_longer_than_time_window(two_layer, basis_matrix.shape[0], axis)

    # convert to array
    two_layer = jax.tree_util.tree_map(jnp.asarray, two_layer)

    # convolve
    conv = _convolve_pad_and_shift(
        basis_matrix,
        two_layer,
        predictor_causality=predictor_causality,
        axis=axis,
        shift=shift,
    )

    #  concatenate back
    flat_stack = [jnp.concatenate(x, axis=axis) for x in conv]

    # re-attach time axis
    flat_stack = [
        type_casting.cast_to_pynapple(x, *time_info[i]) if is_nap[i] else x
        for i, x in enumerate(flat_stack)
    ]

    # recreate tree
    return jax.tree_util.tree_unflatten(struct, flat_stack)
